{% extends "page.html" %}
{% if announcement_home is string %}
  {% set announcement = announcement_home %}
{% endif %}
{% block main %}
  <div class="container">
    <h1 class="sr-only">JupyterHub home page</h1>
    <div class="row">
      <div class="text-center">
        {% if default_server.active %}<a id="stop" role="button" class="btn btn-lg btn-danger">Stop My Server</a>{% endif %}
        <a id="start"
           role="button"
           class="btn btn-lg btn-primary"
           href="{{ url }}">
          {% if not default_server.active %}Start{% endif %}
          My Server
        </a>
      </div>
    </div>
    {% if allow_named_servers %}
      <h2>Named Servers</h2>
      <p>
        In addition to your default server,
        you may have additional
        {% if named_server_limit_per_user > 0 %}{{ named_server_limit_per_user }}{% endif %}
        server(s) with names.
        This allows you to have more than one server running at the same time.
      </p>
      {% set named_spawners = user.all_spawners(include_default=False)|list %}
      <table class="server-table table table-striped">
        <thead>
          <tr>
            <th>Server name</th>
            <th>URL</th>
            <th>Last activity</th>
            <th>Actions</th>
          </tr>
        </thead>
        <tbody>
          <tr class="home-server-row add-server-row">
            <td colspan="4">
              <div class="input-group">
                <input class="new-server-name form-control"
                       aria-label="server name"
                       placeholder="name-your-server">
                <button role="button"
                        type="button"
                        class="new-server-btn btn btn-xs btn-primary">Add New Server</button>
              </div>
            </td>
          </tr>
          {% for spawner in named_spawners %}
            <tr class="home-server-row" data-server-name="{{ spawner.name }}">
              {# name #}
              <td>{{ spawner.name }}</td>
              {# url #}
              <td>
                <a class="server-link {% if not spawner.ready %}hidden{% endif %}"
                   href="{{ user.server_url(spawner.name) }}">{{ user.server_url(spawner.name) }}</a>
              </td>
              {# activity #}
              <td class='time-col'>
                {% if spawner.last_activity %}
                  {{ spawner.last_activity.isoformat() + 'Z' }}
                {% else %}
                  Never
                {% endif %}
              </td>
              {# actions #}
              <td>
                <a role="button"
                   class="stop-server btn btn-xs btn-danger{% if not spawner.active %} hidden{% endif %}"
                   id="stop-{{ spawner.name }}">stop</a>
                <a role="button"
                   class="start-server btn btn-xs btn-primary {% if spawner.active %}hidden{% endif %}"
                   id="start-{{ spawner.name }}"
                   href="{{ base_url }}spawn/{{ user.name }}/{{ spawner.name }}">start</a>
                <button role="button"
                        class="delete-server btn btn-xs btn-danger{% if spawner.active %} hidden{% endif %}"
                        id="delete-{{ spawner.name }}">delete</button>
              </td>
            </tr>
          {% endfor %}
        </tbody>
      </table>
    {% endif %}
  </div>
{% endblock main %}
{% block script %}
  {{ super() }}
  <script type="text/javascript">
    require(["home"]);
  </script>
{% endblock script %}
