{% extends "page.html" %}
{% block main %}
  <div id="react-admin-hook">
    <script id="jupyterhub-admin-config">
      window.api_page_limit = parseInt("{{ api_page_limit|safe }}")
      window.base_url = "{{ base_url|safe }}"
    </script>
    <script src={{ static_url("js/admin-react.js") }}></script>
  </div>
{% endblock main %}
{% block footer %}
  <div class="py-2 px-4 bg-body-tertiary small version_footer">JupyterHub {{ server_version }}</div>
{% endblock footer %}
