{% extends "templates/page.html" %}
{% block announcement %}
  <div class="container text-center announcement"></div>
{% endblock announcement %}
{% block script %}
  {{ super() }}
  <script>
    $.get("/services/announcement/", function(data) {
      $(".announcement").html(data["announcement"]);
    });
  </script>
{% endblock script %}
